granData->contactPointGeometryA[myContactID] = locCPA;
granData->contactPointGeometryB[myContactID] = locCPB;
granData->contactForces[myContactID] = force;
granData->contactTorque_convToForce[myContactID] = torque_only_force;
