// nJitifiableClumpComponents elements are in these arrays
__constant__ __device__ float Radii[] = {_Radii_};
__constant__ __device__ float CDRelPosX[] = {_CDRelPosX_};
__constant__ __device__ float CDRelPosY[] = {_CDRelPosY_};
__constant__ __device__ float CDRelPosZ[] = {_CDRelPosZ_};