deme::clumpComponentOffset_t myCompOffset = granData->clumpComponentOffset[sphereID];
myRelPos.x = CDRelPosX[myCompOffset];
myRelPos.y = CDRelPosY[myCompOffset];
myRelPos.z = CDRelPosZ[myCompOffset];
myRadius = Radii[myCompOffset];
