// nAnalGM elements are in these arrays
__constant__ __device__ deme::objType_t objType[] = {_objType_};
__constant__ __device__ deme::bodyID_t objOwner[] = {_objOwner_};
__constant__ __device__ float objNormal[] = {_objNormal_};
__constant__ __device__ deme::materialsOffset_t objMaterial[] = {_objMaterial_};
__constant__ __device__ float objRelPosX[] = {_objRelPosX_};
__constant__ __device__ float objRelPosY[] = {_objRelPosY_};
__constant__ __device__ float objRelPosZ[] = {_objRelPosZ_};
__constant__ __device__ float objRotX[] = {_objRotX_};
__constant__ __device__ float objRotY[] = {_objRotY_};
__constant__ __device__ float objRotZ[] = {_objRotZ_};
__constant__ __device__ float objSize1[] = {_objSize1_};
__constant__ __device__ float objSize2[] = {_objSize2_};
__constant__ __device__ float objSize3[] = {_objSize3_};
__constant__ __device__ float objMass[] = {_objMass_};