// nDistinctMassProperties elements are in these arrays
__constant__ __device__ float moiX[] = {_moiX_};
__constant__ __device__ float moiY[] = {_moiY_};
__constant__ __device__ float moiZ[] = {_moiZ_};